import random

from colored import fore, back, style

from .monster import Monster
from .room_list import RoomList

class Room:
    def __init__(self, location, data = None, monsterData = None):
        self.seen = False
        self.known = False
        self.hasContents = False
        self.monster = None
        self.trap = None
        self.stairs = ""
        if data:
            for k in data:
                setattr(self, k, data[k])
            if monsterData:
                self.monster = Monster(monsterData["level"], monsterData)
            else:
                self.monster = None
        else:
            self.generateName()
            self.hasContents = self.known

    def generateContents(self, dungeonLevel, floor = -1):
        if not self.hasContents:
            self.hasContents = True
            self.known = True
            if floor > -1: # boss monster
                self.monster = Monster(dungeonLevel - 1, { "floor": floor, "id": -1 })
            else:
                self.monster = Monster(dungeonLevel - 1)

    def generateName(self):
        descriptors = []
        descriptorKeys = random.sample(RoomList.descriptor_types, random.randint(1, 2))

        for key in RoomList.descriptor_types:
            if key in descriptorKeys:
                descriptors.append(random.choice(RoomList.descriptors[key]))

        self.name = f"{fore.CYAN}A {' '.join(descriptors)} room"

    def printStats(self, exits):
        print(self.name + style.RESET)
        doors = []
        stairs = ""
        for exitDir, exitObj in exits:
            if exitObj.isValid():
                if exitDir == "n":
                    doors.append("North")
                if exitDir == "s":
                    doors.append("South")
                if exitDir == "e":
                    doors.append("East")
                if exitDir == "w":
                    doors.append("West")
                if exitDir == "up":
                    stairs = "There are stairs here going up."
                if exitDir == "down":
                    stairs = "There are stairs here going down."
        if len(doors) == 1:
            print(f"There is an exit to the {doors[0]}.")
        else:
            sliceObj = slice(len(doors) - 1)
            print(f"There are exits to the {', '.join(doors[sliceObj])} and {doors[len(doors) - 1]}.")
        if stairs:
            print(stairs)

    def getMapIcon(self, monsterList):
        out = " "
        if self.monster and self.monster.known:
            monsterList.append(self.monster)
            out = f"{back.DARK_RED_1}{fore.ORANGE_3}{len(monsterList)}{style.RESET}"
        return out

    def removeMonster(self):
        self.monster = None

    def printWall(self, side, stairs):
        if self.known:
            glyph = '[' if side == "left" else ']'
        else:
            glyph = ' '
        
        wallColor = style.RESET
        wallStyle = ""
        if stairs:
            wallColor = fore.RED if stairs == "down" else fore.GREEN
            wallStyle = style.BOLD
        if not self.seen:
            wallStyle = style.DIM

        return f"{wallColor}{wallStyle}{glyph}{style.RESET}"

    def printMap(self, isCurrentRoom, monsterList, stairs = ""):
        out = ""
        out += self.printWall("left", stairs)
        out += f"{fore.WHITE}{back.BLUE}{style.BOLD}*{style.RESET}" if isCurrentRoom else self.getMapIcon(monsterList)
        out += self.printWall("right", stairs)
        return out


import random
import math

from colored import fore, back, style

from .item_list import ItemList
from .utils import Utils

class Item:
    def __init__(self, level, data = None, force = []):
        self.isEgo = False
        self.equipped = False
        self.kind = ''
        self.stack = 1
        self.id = 0

        if level > 0:
            self.level = level
        
        if data:
            for i in data:
                setattr(self, i, data[i])
        else:
            info = self.getItem(level)
            while len(force) > 0 and (not info or info["kind"] not in force):
                info = self.getItem(level // 2)
            if info != None:
                for i in info:
                    setattr(self, i, info[i])
                
                if self.effect:
                    self.displayName = f"{self.name} of {self.effect.title()}"
                elif self.ability:
                    self.displayName = f"{self.name} of {self.ability.title()}"
                else:
                    self.displayName = self.name

                # improved items
                if self.kind == "ring":
                    self.level = min((level + 1) // 2, 10)
                    descriptor = ItemList.descriptors["ring"][self.level]
                elif self.kind in ["weapon", "armor"]:
                    levelDiff = level - self.level
                    if levelDiff > 0:
                        try:
                            descriptor = ItemList.descriptors[self.kind][levelDiff // 2]
                        except KeyError:
                            pass
                        try:
                            descriptor = ItemList.descriptors[self.kind][self.type][levelDiff // 2]
                        except KeyError:
                            pass
                        except TypeError:
                            pass
                        self.displayName = f"{descriptor} {self.displayName}"
                        self.level += levelDiff
                        if self.atk:
                            self.atk += levelDiff // 2
                        if self.ac:
                            self.ac += levelDiff // 2

                if self.kind in ["weapon", "armor"]:
                    egoChance = random.randint(1,100)
                    if egoChance <= level:
                        self.generateEgo()

    def generateEgo(self):
        self.isEgo = True

        if self.kind == "weapon":
            ego = random.choice(ItemList.weaponEgo)
            self.atk += ego["bonus"]
        else:
            ego = random.choice(ItemList.armorEgo)
            self.ac += ego["bonus"]

        self.displayName += f" of {ego['name']}"
        for key in ego["attributes"]:
            setattr(self, key, ego["attributes"][key])

    def getAbilityLevel(self):
        if self.ability:
            return self.level if self.kind == "ring" else (self.level + 1) // 2
        else:
            return 0

    def getPrice(self, valueFactor):
        price = self.level * valueFactor
        if self.kind == "usable":
            price = price * 2
        if self.isEgo:
            price = price * 3
        return price

    def displayStack(self):
        return f" [{self.stack}]" if self.stack > 1 else ""

    def displayBonus(self):
        out = ""
        if self.atk:
            out += f"{self.atk}"
        if self.ac:
            out += f"{self.ac}"
        if self.ability and not "resist_" in self.ability:
            ability = self.getAbilityLevel()
            out += f"{ability}" if out == "" else f"({ability})"
        return out

    def displayAbility(self):
        out = ""
        abbreviations = {
            "regeneration": "regen",
            "resist_fire": "res. fire",
            "resist_cold": "res. cold",
            "resist_acid": "res. acid",
            "resist_electric": "res. electric"
        }
        if self.ability:
            try:
                out += abbreviations[self.ability]
            except KeyError:
                out += self.ability
        return out

    @staticmethod
    def getItem(level):
        itemRoll = random.randint(level, 100)
        kind = "none"
        
        if itemRoll < 80:
            return None
        if itemRoll < 88:
            kind = "usable"
        elif itemRoll < 93:
            kind = "weapon"
        elif itemRoll < 98:
            kind = "armor"
        else:
            kind = "ring"
    
        if kind in ["usable", "ring"]:
            items = [item for item in ItemList.items if item["kind"] == kind and item["level"] <= level] 
        else:
            items = [item for item in ItemList.items if item["kind"] == kind and item["level"] <= level and item["level"] > level - 8] 
        return random.choice(items) if len(items) else None

    @staticmethod
    def getOptions(source, options):
        page = options["currPage"]
        pageSize = options["pageSize"]
        filterValue = options["filter"]
        mode = options["mode"]
        message = options["message"]
        if mode in ["buy", "sell"]:
            message = f"{fore.YELLOW}GP: {options['gp']}  {style.RESET}{message}"
        sourceType = type(source).__name__
        actions = []
        if sourceType == "Player":
            if mode == "combat":
                actions = ["<U>se"]  
            elif mode == "sell":
                actions = ["<B>uy", "<S>ell"]  
            else:
                actions = ["<E>quip", "<U>se"]
        elif sourceType == "Store":
            actions = ["<B>uy", "<S>ell"]
        
        filteredItems = source.items if filterValue == "all" else list(filter(lambda i: i.kind == filterValue, source.items))
        totalItems = len(filteredItems)

        navigation = []
        if page > 0:
            navigation += ["<P>rev Page"]
        if (page + 1) * pageSize < totalItems:
            navigation += ["<N>ext Page"]

        leave = ["<L>eave"] if sourceType == "Store" else ["<C>lose"]
        filterOption = [] if mode == "combat" else ["<F>ilter"]
        out = actions + navigation + filterOption + leave
        out = message + f"{style.RESET}\n" + ", ".join(out) + style.RESET
        options["message"] = ""
        return out


    @staticmethod
    def printInventory(source, options):
        page = options["currPage"]
        pageSize = options["pageSize"]
        filterValue = options["filter"]
        sourceType = type(source).__name__
 
        filterHeader = "" if filterValue == "all" else f" ({filterValue.title()})"
        if sourceType == "Player":
            print(f"{fore.CYAN}{style.BOLD}{source.name}'s Inventory{filterHeader}{style.RESET}")
            valueFactor = options["sellFactor"]
        elif sourceType == "Store":
            print(f"{fore.MAGENTA}{style.BOLD}Store Inventory{filterHeader}{style.RESET}")
            valueFactor = options["buyFactor"]

        filteredItems = source.items if filterValue == "all" else list(filter(lambda i: i.kind == filterValue, source.items))

        startIndex = pageSize * page
        s = slice(startIndex, startIndex + pageSize)
        pageItems = filteredItems[s]
        
        itemLines = []
        for i, item in enumerate(pageItems, 1):
            line = {
                "Name": f"{startIndex + i}) {item.displayName}{item.displayStack()}",
                "Bonus": item.displayBonus(),
                "Type": f"{item.type}",
                "Ability": item.displayAbility(),
                "Value": f"{item.getPrice(valueFactor)}"
            }
            if sourceType == "Player":
                line["_color"] = fore.GREEN if item.equipped else style.RESET

            itemLines.append(line)

        Utils.printTable(["   Name", "Bonus", "Type", "Ability", "Value"], itemLines, [40, 7, 8, 12, 8])

    @staticmethod
    def getFilteredItem(source, options, index):
        filterValue = options["filter"]
 
        filteredItems = source.items if filterValue == "all" else list(filter(lambda i: i.kind == filterValue, source.items))

        try:
            return filteredItems[index]
        except IndexError:
            return None


import random
import copy
import sys

from colored import fore, back, style

from .room import Room
from .door import Door

class Map:
    def __init__(self, numFloors = 10, width = 10, data = None):
        self.floors = []
        self.width = width
        self.numFloors = numFloors
        self.dungeonLevel = 1 if not data else data["dungeonLevel"]
        self.message = ""

        for f in range(numFloors):
            floor = {
                "rooms": {},
                "doors": {
                    "ew": {},
                    "ns": {}
                },
                "stairs": {}
            }

            if data:
                for r in range(width):
                    for c in range(width):
                        # generate rooms
                        floor["rooms"][(r,c)] = Room((f, r, c), data["rooms"][f"{f}-{r}-{c}"], data["monsters"][f"{f}-{r}-{c}"])
                        # load east-west doors
                        if c < width - 1:
                            floor["doors"]["ew"][(r,c)] = Door("ew", data["doors"]["ew"][f"{f}-{r}-{c}"])
                        # load north-south doors
                        if r < width - 1:
                            floor["doors"]["ns"][(r,c)] = Door("ns", data["doors"]["ns"][f"{f}-{r}-{c}"])
                        # load stairs
                        try: 
                            floor["stairs"][(r,c)] = Door("stairs", data["stairs"][f"{f}-{r}-{c}"])
                        except KeyError:
                            pass
            else:
                for r in range(width):
                    for c in range(width):
                        # generate rooms
                        floor["rooms"][(r,c)] = Room((f, r, c))
                        # generate east-west doors
                        if c < width - 1:
                            floor["doors"]["ew"][(r,c)] = Door("ew")
                        # generate north-south doors
                        if r < width - 1:
                            floor["doors"]["ns"][(r,c)] = Door("ns")

                # connect rooms and doors
                for r in range(width):
                    for c in range(width):
                        doorList = []
                        if r > 0:
                            doorList.append(floor["doors"]["ns"][(r-1,c)])
                        if r < self.width - 1:
                            doorList.append(floor["doors"]["ns"][(r,c)])
                        if c > 0:
                            doorList.append(floor["doors"]["ew"][(r,c-1)])
                        if c < self.width - 1:
                            doorList.append(floor["doors"]["ew"][(r,c)])

                        # check for disconnected room and fix it
                        connected = False
                        for d in doorList:
                            if d.exists:
                                connected = True
                        if not connected:
                            random.choice(doorList).exists = True
            
                # generate stairs
                up = (-1, -1)
                if f > 0:
                    for i, (key, stair) in enumerate(self.floors[f - 1]["stairs"].items()):
                        if stair.stairDir == "down":
                            up = key
                            upStairs = copy.deepcopy(stair)
                            upStairs.stairDir = "up"
                            floor["stairs"][key] = upStairs
                            floor["rooms"][key].stairs = "up"
                if f < numFloors:
                    down = (random.randint(0,self.width - 1), random.randint(0,self.width - 1))
                    while down == up:
                        down = (random.randint(0,self.width - 1), random.randint(0,self.width - 1))
                    downStairs = Door("stairs")
                    downStairs.stairDir = "down"
                    floor["stairs"][down] = downStairs
                    floor["rooms"][down].stairs = "down"

            self.floors.append(floor)

        if data:
            self.playerPosition = data["playerPosition"]
        else:
            self.setPlayerPosition(0, -1, -1)
            escapeStairs = Door("stairs")
            escapeStairs.stairDir = "up"
            self.floors[0]["stairs"][(self.playerPosition[1], self.playerPosition[2])] = escapeStairs

    def getCurrentRoom(self):
        pp = self.playerPosition
        return self.floors[pp[0]]["rooms"][(pp[1],pp[2])]

    def getRoom(self, floor, row, col):
        return self.floors[floor]["rooms"][(row, col)]

    def movePlayer(self, direction):
        newPos = list(self.playerPosition)
        if direction == "n":
            newPos[1] -= 1
        elif direction == "s":
            newPos[1] += 1
        elif direction == "e":
            newPos[2] += 1
        elif direction == "w":
            newPos[2] -= 1
        elif direction == "u":
            newPos[0] -= 1
        elif direction == "d":
            newPos[0] += 1
        
        self.setPlayerPosition(newPos[0], newPos[1], newPos[2])

    def setPlayerPosition(self, floor, row, col):
        startRoom = False
        if row == -1 or col == -1:
            startRoom = True
            row = random.randint(0, self.width -1)
            col = random.randint(0, self.width - 1)
        self.playerPosition = [floor, row, col]
        
        room = self.getRoom(floor, row, col)
        room.seen = True
        if room.monster:
            room.monster.seen = True
            room.monster.known = True
        room.known = True

        if startRoom:
            room.hasContents = True
            room.monster = None
        if not room.hasContents:
            self.fillRoom(room)
        
        for key, door in self.getCurrentDoors():
            door.seen = True
            if (door.exists):
                if key == "n":
                    self.fillRoom(self.getRoom(floor, row - 1, col))
                if key == "s":
                    self.fillRoom(self.getRoom(floor, row + 1, col))
                if key == "e":
                    self.fillRoom(self.getRoom(floor, row, col + 1))
                if key == "w":
                    self.fillRoom(self.getRoom(floor, row, col - 1))

    def resetRooms(self):
        for f in range(self.numFloors):
            floor = self.floors[f]
            for r in range(self.width):
                for c in range(self.width):
                    room = floor["rooms"][(r,c)]
                    room.hasContents = False
                    room.monster = None
        pp = self.playerPosition
        lastRoom = self.getRoom(pp[0], pp[1], pp[2])
        lastRoom.hasContents = True
        self.setPlayerPosition(pp[0], pp[1], pp[2])

    def fillRoom(self, room):
        if room.stairs == "down":
            room.generateContents(self.dungeonLevel, self.playerPosition[0])
        else:
            room.generateContents(self.dungeonLevel)

    def printFloor(self, turn):
        floor = self.playerPosition[0]
        mapBuffer = []
        legendBuffer = [
            f"{fore.CYAN}{style.BOLD}Legend:{style.RESET}", 
            f"{fore.GREEN}[ ]{style.RESET} / {fore.RED}[ ]{style.RESET} : Stairs Up/Down"
        ]

        ppSlug = f"[{self.playerPosition[1]},{self.playerPosition[2]}]"
        print(f"{fore.MAGENTA}{style.BOLD}Dungeon Level {floor + 1} - Turn {turn}{style.RESET}")

        # print top map header
        header = f"{style.DIM}  "
        for c in range(self.width):
            header +=(f" {c} ")
            if (c < self.width - 1):
                header += "  "
        mapBuffer.append(header)

        # print map rows
        monsterList = []
        for r in range(self.width):
            roomRow = ""
            roomRow += f"{style.DIM}{r} {style.RESET}"
            for c in range(self.width):
                room = self.floors[floor]["rooms"][(r,c)]
                isCurrentRoom = self.playerPosition[1] == r and self.playerPosition[2] == c
                stairType = ""
                try:
                    stairs = self.floors[floor]["stairs"][(r,c)]
                    stairType = stairs.stairDir
                except KeyError:
                    pass
                roomRow += room.printMap(isCurrentRoom, monsterList, stairType)
                if c < self.width - 1:
                    door = self.floors[floor]["doors"]["ew"][(r,c)]
                    roomRow += door.printMap()
            mapBuffer.append(roomRow)

            if r < self.width - 1:
                doorRow = "  "
                for c in range(self.width):
                    door = self.floors[floor]["doors"]["ns"][(r,c)]
                    doorRow += door.printMap()
                    if c < self.width - 1:
                        doorRow += "  "
                mapBuffer.append(doorRow)
        
        for i, monster in enumerate(monsterList):
            monsterLine = f"{back.DARK_RED_1}{fore.ORANGE_3}{i + 1}{style.RESET} {monster.name}"
            if monster.seen:
                monsterLine += f" {style.DIM}({monster.hp}/{monster.maxHp})"
            legendBuffer.append(monsterLine)

        mapLineWidth = len(mapBuffer[0])
        for b in range(max(len(mapBuffer), len(legendBuffer))):
            mapLine = " " * mapLineWidth
            legendLine = ""
            try:
                mapLine = mapBuffer[b]
            except IndexError:
                pass
            try:
                legendLine = legendBuffer[b]
            except IndexError:
                pass
            print(f"{mapLine} {style.DIM}| {style.RESET}{legendLine}")

    def getCurrentDoors(self):
        floor = self.floors[self.playerPosition[0]]
        r = self.playerPosition[1]
        c = self.playerPosition[2]
        return self.getRoomDoors(floor, r, c)

    def getRoomDoors(self, floor, r, c):
        out = []
        if r > 0:
            out.append(("n", floor["doors"]["ns"][(r-1,c)]))
        if r < self.width - 1:
            out.append(("s", floor["doors"]["ns"][(r,c)]))
        if c > 0:
            out.append(("w", floor["doors"]["ew"][(r,c-1)]))
        if c < self.width - 1:
            out.append(("e", floor["doors"]["ew"][(r,c)]))
        for key, stair in floor["stairs"].items():
            if key == (r, c):
                out.append((stair.stairDir, stair))  
        return out

    def getConnectedRooms(self):
        pp = self.playerPosition
        doors = self.getCurrentDoors()
        out = []
        for (direction, door) in doors:
            if door.exists:
                if direction == "n":
                    out.append(self.getRoom(pp[0], pp[1] - 1, pp[2]))
                elif direction == "s":
                    out.append(self.getRoom(pp[0], pp[1] + 1, pp[2]))
                elif direction == "e":
                    out.append(self.getRoom(pp[0], pp[1], pp[2] + 1))
                elif direction == "w":
                    out.append(self.getRoom(pp[0], pp[1], pp[2] - 1))
        return out

    def discoverStairs(self):
        pp = self.playerPosition
        floor = self.floors[pp[0]]
        for key, stair in floor["stairs"].items():
            if stair.stairDir == "down":
                room = self.getRoom(pp[0], key[0], key[1])
                self.fillRoom(room)
                return

    def getOptions(self):
        out = f"{style.BOLD}"
        options = []
        doors = []
        stairs = []
        for direction, door in self.getCurrentDoors():
            if (door.isValid()):
                if direction == "n":
                    doors += ["<N>orth"]
                if direction == "s":
                    doors += ["<S>outh"]
                if direction == "e":
                    doors += ["<E>ast"]
                if direction == "w":
                    doors += ["<W>est"]
                if direction == "up":
                    stairs += ["<U>p"]
                if direction == "down":
                    stairs += ["<D>own"]
        options += doors + stairs + ["<L>isten", "<S>earch", "<B>ack"]
        out = self.message + f"{style.RESET}\n" + ", ".join(options) + style.RESET
        self.message = ""
        return out




import random
import math
import os
import json
import copy
import sys
import time
from types import *

from colored import fore, back, style

from .monster import Monster
from .item import Item
from .player import Player
from .room import Room
from .maps import Map
from .door import Door
from .store import Store
from .utils import Utils

clear=lambda: os.system('cls' if os.name == 'nt' else 'clear')

class Game:
    saveFilePath = os.path.join(os.path.abspath(os.path.dirname(__file__)), "save")
    saveListFilePath = os.path.join(saveFilePath, "saveList.json")
    playerQuit = False
    restart = False

    def __init__(self):
        self.initialize()
        
    def initialize(self):
        self.mode = "peace"
        self.restart = False
        self.turn = 1
        self.nextLevel = 100
        self.level = 1
        self.monster = None
        self.map = None
        self.store = None
        self.saveSlot = -1
        self.options = {
            "combat": "\n<A>ttack, <D>efend, <X>amine, <U>se Item, <R>un",
            "peace": "\n<R>est, <C>ontinue, <I>nventory, <M>erchant, <S>ave, <Q>uit",
            "gameOver": "\n<R>estart, <Q>uit",
            "inventory": lambda : Item.getOptions(self.player, self.itemListOptions) if self.player else "",
            "store": lambda : Item.getOptions(self.store, self.itemListOptions) if self.store else "",
            "map": lambda : self.map.getOptions() if self.map else ""
        }
        self.resolver = {
            "combat": self.combatResolve,
            "peace": self.peaceResolve,
            "gameOver": self.gameOverResolve,
            "inventory": self.inventoryResolve,
            "store": self.storeResolve,
            "map": self.mapResolve
        }
        self.display = {
            "combat": self.combatDisplay,
            "peace": self.peaceDisplay,
            "gameOver": self.gameOverDisplay,
            "inventory": self.inventoryDisplay,
            "store": self.storeDisplay,
            "map": self.mapDisplay
        }
        self.effectResolvers = {
            "healing": self.healingResolve,
            "time warp": self.timeWarpResolve,
            "shopping": self.shoppingResolve,
            "blinking": self.blinkingResolve,
            "fireball": self.fireballResolve,
            "digging": self.diggingResolve,
            "teleport": self.teleportResolve,
            "invisibility": self.invisibilityResolve,
            "knowledge": self.knowledgeResolve
        }
        self.inititemListOptions()
        self.clearResolution()

    def inititemListOptions(self):
        self.itemListOptions = {
            "currPage": 0,
            "pageSize": 15,
            "filter": "all",
            "mode": "peace",
            "sellFactor": 20,
            "buyFactor": 100,
            "message": "",
            "gp": 0
        }

### NUMBER METHODS ###

    def rollDie(self, size):
        return random.randint(1, size)
    
    def rollDamage(self, creature):
        return self.rollDie(creature.level) + self.rollDie(creature.level) + self.rollDie(creature.atk)
    
    def incrementDungeonLevel(self):
        self.level += 1
        self.nextLevel = self.turn + (100 * self.level)
        self.map.dungeonLevel = self.level

    def incrementTurn(self, numTurns = 1):
        self.turn += numTurns
        if self.turn >= self.nextLevel:
            self.incrementDungeonLevel()
            self.addResolution(f"{fore.RED}The dungeon seems more dangerous...")
    
### PRINTING METHODS ###

    def printRoll(self, roll, target):
        self.addResolution(f"[{roll} vs {target}]")
        
    def printStats(self):
        displayFunc = self.display[self.mode]
        displayFunc()

    def printStatHeader(self):
        print(f"{style.BOLD}Dungeon Level {self.map.playerPosition[0] + 1} - Turn {self.turn}{style.DIM}/{self.nextLevel}\n{style.RESET}")

    def clearResolution(self):
        self.resolution = []

    def addResolution(self, text):
        self.resolution.append(text)

    def printResolution(self):
        if len(self.resolution) > 0:
            print("\n" + f"{style.RESET}\n".join(self.resolution) + style.RESET)

### DISPLAY METHODS ###

    def combatDisplay(self):
        self.printStatHeader()
        self.player.printStats()
        print("")
        self.monster.printStats(self.player.monsterLore)

    def peaceDisplay(self):
        self.printStatHeader()
        self.player.printStats()
        print("")
        room = self.map.getCurrentRoom()
        room.printStats(self.map.getCurrentDoors())

    def inventoryDisplay(self):
        Item.printInventory(self.player, self.itemListOptions)

    def storeDisplay(self):
        invSource = self.store if self.itemListOptions["mode"] == "buy" else self.player
        self.itemListOptions["gp"] = self.player.gp
        Item.printInventory(invSource, self.itemListOptions)

    def mapDisplay(self):
        self.map.printFloor(self.turn)

    def gameOverDisplay(self):
        self.player.printHistory(self.turn)

    def printOptions(self):
        options = self.options[self.mode]
        if type(options) is LambdaType:
            options = options()
        print(f"{style.BOLD}{options}")

### COMBAT METHODS ###

    def playerAttack(self):
        if self.monster.charges > 0:
            self.player.incrementHistory("reckless")

        atkRoll = self.rollDie(20) + self.player.atk
        
        monsterDefense = self.monster.ac
            
        if atkRoll == 20 or atkRoll >= monsterDefense:
            damRoll = self.rollDamage(self.player)
            damage = self.monster.damage(damRoll, self.player.atkType)
            self.addResolution(f"{fore.GREEN}You {self.player.getAtkVerb()} the {self.monster.name} for {damage}")
            if damage < damRoll:
                self.addResolution(f"{fore.MAGENTA}Your weapon is not very effective.")
                self.addLore("resist")
            if damage > damRoll:
                self.addResolution(f"{fore.CYAN}Your weapon is very effective.")
                self.addLore("vulnerability")
            self.player.incrementHistory("dmg_done", damage)
        else:
            self.addResolution(f"{fore.WHITE}You miss!")
                
    def monsterAttack(self, playerDefending):
        if self.monster.quotes:
            quoteRoll = self.rollDie(3)
            if quoteRoll == 1:
                self.addResolution(f"{fore.LIGHT_SLATE_BLUE}{random.choice(self.monster.quotes)}")

        if self.monster.charges == 0:
            chargeRoll = self.rollDie(10)
            if chargeRoll == 1:
                self.monster.charges += self.monster.chargeRate
                self.addResolution(f"{fore.DARK_ORANGE_3B}The {self.monster.name} readies an attack...")
                return

        atkRoll = self.rollDie(20) + self.monster.atk
        
        playerDefense = self.player.ac
        if playerDefending:
            playerDefense = playerDefense * 2 - 10

        if atkRoll == 20 or atkRoll >= playerDefense:
            damRoll = self.rollDamage(self.monster)
            self.addResolution(f"{fore.RED}The {self.monster.name} {self.monster.getAtkVerb()} you for {style.BOLD}{damRoll}")
            self.player.damage(damRoll, self.monster.atk_type)
            self.player.incrementHistory("dmg_taken", damRoll)
        else:
            self.addResolution(f"{fore.WHITE}The {self.monster.name} misses!")

        while self.monster.charges > 0:
            self.monster.charges -= 1
            specAtkRoll = self.rollDie(20) + self.monster.atk
            if atkRoll >= playerDefense:
                self.addLore("special")
                if self.monster.special == "drain":
                    self.player.drain(random.randint(1, self.monster.level) * 50)
                    self.addResolution(f"{fore.PALE_TURQUOISE_1}It drains your life essence!")
                elif self.monster.special == "melt":
                    if "acid" in self.player.resist:
                        self.addResolution(f"{fore.GREEN}You resist the spray of acid!")
                    else:
                        meltables = [item for item in self.player.items if item.kind == "weapon" or item.type == "metal"]
                        if len(meltables) > 0:
                            item = random.choice(meltables)
                            self.addResolution(f"{fore.CHARTREUSE_1}It melts your {item.displayName} with acid!")
                            self.player.removeItem(item)
                elif self.monster.special == "burn":
                    if "fire" in self.player.resist:
                        self.addResolution(f"{fore.GREEN}You resist the burst of flame!")
                    else:
                        burnables = [item for item in self.player.items if item.name == "Scroll" or item.type in ["cloth", "leather"]]
                        if len(burnables) > 0:
                            item = random.choice(burnables)
                            self.addResolution(f"{fore.CHARTREUSE_1}It burns your {item.displayName} to ash!")
                            self.player.removeItem(item)
                elif self.monster.special == "freeze":
                    if "cold" in self.player.resist:
                        self.addResolution(f"{fore.GREEN}You resist the blast of frost!")
                    else:
                        freezables = [item for item in self.player.items if item.name == "Potion"]
                        if len(freezables) > 0:
                            item = random.choice(freezables)
                            self.addResolution(f"{fore.CHARTREUSE_1}It freezes your {item.displayName} and shatters it!")
                            self.player.removeItem(item)
                elif self.monster.special == "shock":
                    if "electric" in self.player.resist:
                        self.addResolution(f"{fore.GREEN}You resist the bolt of lightning!")
                    else:
                        self.addResolution(f"{fore.CHARTREUSE_1}It shocks you and makes you drop your weapon!")
                        self.player.unequipItem("weapon")
                else:
                    damRoll = self.rollDamage(self.monster)
                    self.addResolution(f"{fore.RED}The {self.monster.name} {self.monster.getAtkVerb()} you for {style.BOLD}{damRoll}")
                    self.player.damage(damRoll, self.monster.atk_type)
                    self.player.incrementHistory("dmg_taken", damRoll)
            else:
                self.addResolution(f"{fore.WHITE}Its special attack fails!")      
        self.playerDeathCheck()

    def getReward(self, level):
        xp = self.monster.level * 50
        gp = self.monster.level * self.rollDie(10)
        self.player.xp += xp
        self.player.gp += gp
        self.addResolution(f"\n{fore.YELLOW}You got {xp} XP and {gp} GP")
        self.checkPlayerLevelUp()
        
        item = Item(self.monster.level, None, ["weapon", "armor", "ring"] if self.monster.isBoss else [])
        if item.kind:
            self.player.addItem(item)
            self.addResolution(f"{fore.YELLOW}You found: {item.displayName}")

    def checkPlayerLevelUp(self):
        levelUp = self.player.checkLevelUp()
        if levelUp:
            if self.level < self.player.level:
                self.incrementDungeonLevel()
            self.addResolution(f"{fore.YELLOW}You gained a level!")

    def monsterDeathCheck(self):
        if not self.monster:
            return
        if self.monster.hp <= 0:
            self.addResolution(f"{fore.GREEN}The {self.monster.name} dies!")
            self.map.getCurrentRoom().removeMonster()
            self.player.incrementHistory("kills")
            if self.player.hp <= self.player.maxHp * .25:
                self.player.incrementHistory("risky_win")
            self.getReward(self.monster.level)
            self.monster = None
            self.mode = "peace"
        else:
            self.monsterAttack(False)  
            self.playerDeathCheck()
                
    def playerDeathCheck(self):
        if self.player.hp <= 0:
            self.addResolution(f"{fore.RED}{style.BOLD}You die!")
            self.player.killedBy(self.monster, self.map.playerPosition[0] + 1)
            self.mode = "gameOver"

    def playerEscape(self):
        self.monster.charges = 0
        self.monster = None
        exits = [e for e in self.map.getCurrentDoors() if e[1].used]
        if len(exits) == 0:
            exits = self.map.getCurrentDoors()
        door = random.choice(exits)
        self.map.movePlayer(door[0])
        newRoom = self.map.getCurrentRoom()
        self.monster = newRoom.monster
        self.mode = "combat" if self.monster else "peace"

    def addLore(self, area = None):
        id = str(self.monster.id)
        try:
            lore = self.player.monsterLore[id]
        except KeyError:
            lore = { "resist": False, "vulnerability": False, "special": False}
            self.player.monsterLore[id] = lore
            
        learned = False
        if not area:
            unknowns = [key for key, value in lore.items() if value == False]
            if len(unknowns) == 0:
                self.addResolution("There is nothing more for you to learn!")
                return
            else:
                area = random.choice(unknowns)
                learned = True
        
        if not lore[area]:
            learned = True
            lore[area] = True
            self.addResolution(f"{fore.DEEP_PINK_3B}You gain new knowledge!")

        if learned:
            if lore["resist"] and lore["vulnerability"] and lore["special"]:
                xp = self.monster.level * 50
                self.player.xp += xp
                self.addResolution(f"{fore.YELLOW}You gained {xp} XP for mastering your knowledge of this creature!")        
                self.checkPlayerLevelUp()

### ITEM METHODS ###

    def filterItems(self):
        print("Filter: <W>eapons, <A>rmor, <R>ings, <U>sable, <E>verything")
        choice = input()
        filterValue = "all"
        if choice.lower() == "w":
            filterValue = "weapon"
        if choice.lower() == "a":
            filterValue = "armor"
        if choice.lower() == "r":
            filterValue = "ring"
        if choice.lower() == "u":
            filterValue = "usable"
        self.itemListOptions["currPage"] = 0
        self.itemListOptions["filter"] = filterValue

    def selectItem(self, source, prompt):
        sys.stdout.write(f"{prompt} {style.DIM}<Enter> to cancel {style.RESET}")
        complete = False
        while not complete:
            itemNum = input()
            if itemNum == "":
                return None
            else:
                try:
                    i = int(itemNum) - 1
                    try:
                        item = Item.getFilteredItem(source, self.itemListOptions, i)
                        complete = True
                        return item
                    except IndexError:
                        print("Invalid item number")
                except ValueError:
                    #Handle the exception
                    print("Please enter an integer")

    def getItemPrice(self, item, action):
        factor = self.itemListOptions["buyFactor" if action == "buy" else "sellFactor"]
        price = item.level * self.itemListOptions["buyFactor"]
        if item.kind == "usable":
            price = price * 2
        return price

    def resolveItem(self, item, mode):
        if item.kind != "usable":
            return False
        if mode == "combat" and item.type == "peace":
            return False
        if mode == "peace" and item.type == "combat":
            return False

        effect = item.effect
        escaped = False

        try:
            used = self.effectResolvers[effect]()
        except KeyError:
            pass

        if used:
            if self.monster:
                self.mode = "combat"
                self.monsterDeathCheck()
            elif self.mode != "store": 
                self.mode = "peace"
            self.incrementTurn()
        return used

    def healingResolve(self):
        self.player.heal(self.player.maxHp // 2)
        self.addResolution(f"{fore.GREEN}You feel much better!")
        return True

    def timeWarpResolve(self):
        self.nextLevel += self.level * 50
        self.addResolution(f"{fore.GREEN}The dungeon feels less dangerous...")
        return True

    def invisibilityResolve(self):
        self.playerEscape()
        self.addResolution(f"{fore.CYAN}You turn invisible and sneak away!")
        return True

    def knowledgeResolve(self):
        self.map.discoverStairs()
        self.nextLevel -= self.level * 25
        self.addResolution(f"{fore.CYAN}The secrets of the dungeon become clearer!")
        self.addResolution(f"{fore.DARK_ORANGE_3B}The monsters sense your questing mind...")
        if self.turn >= self.nextLevel:
            self.incrementDungeonLevel()
        return True

    def teleportResolve(self):
        pp = self.map.playerPosition
        prompt = "Teleport which direction?"
        validDirs = []
        if pp[1] > 0:
            validDirs += "n"
            prompt += " <N>orth"
        if pp[1] < self.map.width - 1:
            validDirs += "s"
            prompt += " <S>outh"
        if pp[2] < self.map.width - 1:
            validDirs += "e"
            prompt += " <E>ast"
        if pp[2] > 0:
            validDirs += "w"
            prompt += " <W>est"
        prompt += f"{style.DIM} <Enter> to cancel {style.RESET}"
        sys.stdout.write(prompt)
        choice = input()
        if choice and choice in validDirs:
            teleport = {
                "n": lambda: self.map.setPlayerPosition(pp[0], random.randint(0, pp[1] - 1), pp[2]),
                "s": lambda: self.map.setPlayerPosition(pp[0], random.randint(pp[1] + 1, self.map.width - 1), pp[2]),
                "e": lambda: self.map.setPlayerPosition(pp[0], pp[1], random.randint(pp[2] + 1, self.map.width - 1)),
                "w": lambda: self.map.setPlayerPosition(pp[0], pp[1], random.randint(0, pp[2] - 1))
            }
            teleport[choice]()
            self.addResolution(f"{fore.CYAN}You are magically transported!")
            newRoom = self.map.getCurrentRoom()
            newRoom.generateContents(self.level)
            self.monster = newRoom.monster
            if self.monster:
                self.mode = "combat"
            else:
                self.mode = "peace"
            return True
        else: 
            return False

    def shoppingResolve(self):
        self.inititemListOptions()
        self.itemListOptions["mode"] = "buy"
        self.store = Store(self.level)
        self.mode = "store"
        bonus = self.player.level * 10
        self.player.gp += bonus
        self.itemListOptions["message"] = f"{fore.YELLOW}You are transported to the store with an extra {bonus} GP!"
        return True

    def blinkingResolve(self):
        pp = self.map.playerPosition
        minR = max(0, pp[1] - 2)
        maxR = min(self.map.width - 1, pp[1] + 2)
        minC = max(0, pp[2] - 2)
        maxC = min(self.map.width - 1, pp[2] + 2)
        newPos = (random.randint(minR, maxR), random.randint(minC, maxC))
        while newPos == (pp[1], pp[2]):
            newPos = (random.randint(minR, maxR), random.randint(minC, maxC))
        self.map.setPlayerPosition(pp[0], newPos[0], newPos[1])
        self.addResolution(f"{fore.MAGENTA}You are transported to another room!")
        newRoom = self.map.getCurrentRoom()
        newRoom.generateContents(self.level)
        self.monster = newRoom.monster
        if self.monster:
            self.mode = "combat"
        else:
            self.mode = "peace"
        return True

    def diggingResolve(self):
        room = self.map.getCurrentRoom()
        doors = [door for door in self.map.getCurrentDoors() if not door[1].exists]

        if len(doors) == 0:
            self.itemListOptions["message"] = f"{fore.ORANGE}There's nowhere to dig!"
            return False

        prompt = "Dig which direction?"
        for (doorDir, door) in doors:
            if doorDir == "n":
                prompt += " <N>orth"
            elif doorDir == "s":
                prompt += " <S>outh"
            elif doorDir == "e":
                prompt += " <E>ast"
            elif doorDir == "w":
                prompt += " <W>est"
        prompt += f"{style.DIM} <Enter> to cancel {style.RESET}"
        sys.stdout.write(prompt)
        choice = input()
        if choice == "":
            return False
        choice = choice[0].lower()
        for (doorDir, door) in doors:
            if choice == doorDir:
                door.exists = True
                self.mode = "peace"
                self.addResolution(f"{fore.DARK_ORANGE_3B}You dug a new door!")
                return True
        return False
             

    def fireballResolve(self):
        dam = 0
        for i in range(self.player.level):
            dam += self.rollDie(8)
        if self.monster.resist == "fire":
            dam = dam // 2
        if self.monster.vulnerability == "fire":
            dam = dam * 2
        self.monster.damage(dam, "fire")
        self.addResolution(f"{fore.GREEN}The {self.monster.name} is blasted for {dam}!")
        return True

### RESOLUTION METHODS ###

    def mapResolve(self, action):
        if action == "B" or action == "":
            self.mode = "peace"
        elif action in ["N", "S", "E", "W", "U", "D"]:
            direction = action.lower()
            for key, door in self.map.getCurrentDoors():
                if key[0] == direction and door.isValid():
                    if direction == "u" and self.map.playerPosition[0] == 0:
                        print(f"{fore.MAGENTA}{style.BOLD}Are you sure you want to exit the dungeon? <Y>es or <N>o{style.RESET}")
                        choice = input()
                        if choice.lower() == "y":
                            self.mode = "gameOver"
                            epitaph = "Defeated the dungeon!" if self.player.hasIdol else "Fled the dungeon!"
                            self.player.setEpitaph(epitaph)
                            return True
                    elif direction == "d" and self.map.playerPosition[0] == self.map.numFloors - 1:
                        self.addResolution(f"{fore.YELLOW}{style.BOLD}You found the Idol of Onekrum!")
                        self.player.hasIdol = True
                        self.addResolution(f"{fore.RED}{style.BOLD}Enraged monsters fill the dungeon...")
                        self.map.resetRooms()
                        return True
                    else:
                        self.map.movePlayer(direction)
                        door.useDoor()
                        if 'traveling' in self.player.abilities:
                            travelRoll = self.rollDie(10)
                            if travelRoll > self.player.getAbilityLevel('traveling'):
                                self.incrementTurn()
                        else:
                            self.incrementTurn()
                        monster = self.map.getCurrentRoom().monster
                        if monster != None:
                            self.monster = monster
                            self.mode = "combat"
        elif action == "L":            
            loreRating = self.player.getLoreRating(self.map.playerPosition[0])
            rooms = self.map.getConnectedRooms()
            success = False
            for room in rooms:
                if room.monster and not room.monster.known:
                    loreRoll = self.rollDie(100)
                    checkNum = loreRating * 3 if self.player.hasSeenMonster(room.monster) else loreRating
                    if (loreRoll <= checkNum):
                        success = True
                        room.monster.known = True
            if success:
                self.map.message = f"{fore.GREEN}You hear something nearby!"
            else:
                self.map.message = f"{fore.CYAN}You listen carefully but don't hear anything new..."
            self.incrementTurn(3)
        else:
            return False
        return True

    def combatResolve(self, action):
        if action == "A":            
            self.playerAttack()      
            self.monsterDeathCheck()
            self.incrementTurn()
        elif action == "D":
            self.addResolution("You defend yourself!")
            self.monsterAttack(True)
            self.incrementTurn()
        elif action == "X":
            self.addResolution("You study your opponent...")
            self.addLore()
            self.monsterAttack(True)
            self.incrementTurn()
        elif action == "U":
            self.mode = "inventory"
            self.inititemListOptions()
            self.itemListOptions["filter"] = "usable"
            self.itemListOptions["mode"] = "combat"
        elif action == "R":
            self.monster.heal(math.floor(self.monster.hd * self.monster.level / (4 + self.player.getAbilityLevel('running') * .075)))
            self.player.incrementHistory("run_away")
            self.playerEscape()
            if 'running' in self.player.abilities:
                self.incrementTurn(10 - self.player.getAbilityLevel('running'))
                self.addResolution(f"{fore.GREEN}You run away very quickly!")
            else:
                self.incrementTurn(10)
                self.addResolution("You run away!")
        else:
            return False
        return True

    def peaceResolve(self, action):
        if action == "C":
            self.mode = "map"
        elif action == "R":
            self.addResolution("You take some time to recover...")
            restFactor = math.sqrt(self.player.level)
            if 'regeneration' in self.player.abilities:
                restFactor -= 0.075 * self.player.getAbilityLevel('regeneration')
                self.addResolution(f"{fore.GREEN}You heal quickly...")
            timeToRest = math.floor((self.player.maxHp - self.player.hp) * restFactor)
            self.player.heal()
            self.player.incrementHistory("rest")
            self.incrementTurn(timeToRest)
        elif action == "I":
            self.inititemListOptions()
            self.mode = "inventory"
        elif action == "M":
            if self.player.hasIdol:
                self.addResolution(f"{fore.RED}It's too dangerous to go to the store!")
                return True
            self.inititemListOptions()
            self.itemListOptions["mode"] = "buy"
            self.store = Store(self.level)
            self.mode = "store"
            timeToShop = (self.map.playerPosition[0] + 1) * (10 - self.player.getAbilityLevel('traveling'))
            self.itemListOptions["message"] = f"{style.DIM}This trip will take {timeToShop} turns..."
            self.incrementTurn(timeToShop)
        elif action == "S":
            self.createSave()
            print("<C>ontinue or <Q>uit?")
            choice = input()
            if len(choice) > 0 and choice[0].upper() == "Q":
                self.endGame()
        elif action == "Q":
            self.endGame()
        else:
            return False
            
        return True

    def gameOverResolve(self, action):
        if action == "R":
            self.restart = True
        elif action == "Q":
            self.endGame()
        return True

    def inventoryResolve(self, action):
        if action == "E":
            item = self.selectItem(self.player, "Which item do you wish to equip?")
            if item:
                self.player.equipItem(item)
        elif action == "P":
            self.itemListOptions["currPage"] -= 1
        elif action == "N":
            self.itemListOptions["currPage"] += 1
        elif action == "F":
            self.filterItems()
        elif action == "U":
            item = self.selectItem(self.player, "Use which item?")
            if item:
                used = self.resolveItem(item, self.itemListOptions["mode"])
                if used:
                    self.player.removeItem(item)
        elif action == "C" or action == "":
            self.mode = self.itemListOptions["mode"]
        return True

    def storeResolve(self, action):
        if action == "B":
            if self.itemListOptions["mode"] == "sell":
                self.itemListOptions["mode"] = "buy"
            else:
                item = self.selectItem(self.store, "Which item do you wish to buy?")
                if item:
                    price = self.getItemPrice(item, "buy")
                    if price <= self.player.gp:
                        self.player.incrementHistory("buy_item")
                        self.itemListOptions["message"] = f"{fore.CYAN}You bought the {item.displayName}!"
                        self.itemListOptions["currPage"] = 0
                        self.player.removeGold(price)
                        self.store.removeItem(item)
                        self.player.addItem(item)
                    else:
                        self.itemListOptions["message"] = f"{fore.RED}You can't afford the {item.displayName}!"
        elif action == "S":
            if self.itemListOptions["mode"] == "buy":
                self.itemListOptions["mode"] = "sell"
            else:
                item = self.selectItem(self.player, "Which item do you wish to sell?")
                if item:
                    self.player.incrementHistory("sell_item")
                    price = self.getItemPrice(item, "sell")
                    self.itemListOptions["message"] = f"{fore.YELLOW}You sold your {item.displayName}!"
                    self.itemListOptions["currPage"] = 0
                    self.player.addGold(price)
                    self.player.removeItem(item)
                    self.store.addItem(item)
        elif action == "P":
            self.itemListOptions["currPage"] -= 1
        elif action == "N":
            self.itemListOptions["currPage"] += 1
        elif action == "F":
            self.filterItems()
        elif action == "L":
            self.mode = "peace"
        return True

### LIFECYCLE METHODS ###

    def takeInput(self):
        self.printOptions()
        resolved = False
        self.clearResolution()
        action = input()
        if len(action) > 0:
            action = action[0].upper()
        self.resolver[self.mode](action)
        
    def nextTurn(self):
        if not self.playerQuit:
            clear()
            self.printStats()
            self.printResolution()
            self.takeInput()

    def startNewGame(self):
        print("Choose a name:")
        name = input()
        self.player = Player(name)
        self.map = Map()

    def endGame(self):
        self.playerQuit = True

    def checkSavePath(self):
        if not os.path.exists(self.saveFilePath):
            os.makedirs(self.saveFilePath)

    def saveWorker(self, saveObj):
        saveFile = open(os.path.join(self.saveFilePath, self.player.saveId), 'w')
        json.dump(saveObj, saveFile)

    def createSave(self):
        self.checkSavePath()

        if self.player.saveId:
            saveId = self.player.saveId 
        else:
            saveId = str(int(time.time()))
            self.player.saveId = saveId

        sys.stdout.write(f"{style.DIM}.")
        saveObj = {
            "player": dict(copy.deepcopy(self.player).__dict__),
            "map": dict(copy.deepcopy(self.map).__dict__),
            "game": {
                "turn": self.turn,
                "nextLevel": self.nextLevel,
                "level": self.level
            }
        }

        # make player items serializable
        sys.stdout.write(fore.BLUE)
        saveObj["items"] = []
        for item in saveObj["player"]["items"]:
            sys.stdout.write(".")
            saveObj["items"].append(item.__dict__)
        del saveObj["player"]["items"]

        # make map rooms serializable
        saveObj["map"]["rooms"] = {}
        saveObj["map"]["monsters"] = {}
        saveObj["map"]["doors"] = {
            "ns": {},
            "ew": {}
        }
        saveObj["map"]["stairs"] = {}

        sys.stdout.write(fore.GREEN)
        for f, floor in enumerate(saveObj["map"]["floors"]):
            sys.stdout.write(".")
            for r in range(self.map.width):
                for c in range(self.map.width):
                    saveKey = f"{f}-{r}-{c}"
                    roomDict = floor["rooms"][(r,c)].__dict__
                    monster = roomDict["monster"]
                    saveObj["map"]["monsters"][saveKey] = monster.__dict__ if monster != None else None
                    del roomDict["monster"]
                    saveObj["map"]["rooms"][saveKey] = roomDict
                    try: 
                        ns = floor["doors"]["ns"][(r,c)].__dict__
                        saveObj["map"]["doors"]["ns"][saveKey] = ns
                    except KeyError:
                        pass
                    try:
                        ew = floor["doors"]["ew"][(r,c)].__dict__
                        saveObj["map"]["doors"]["ew"][saveKey] = ew
                    except KeyError:
                        pass
                    try:
                        stairs = floor["stairs"][(r,c)].__dict__
                        saveObj["map"]["stairs"][saveKey] = stairs
                    except KeyError:
                        pass

        sys.stdout.write(f"\n{style.RESET}")
        del saveObj["map"]["floors"]

        self.saveWorker(saveObj)

        try:
            with open(self.saveListFilePath) as json_file:  
                saveList = json.load(json_file)
        except:
            saveList = {}

        saveList[saveId] = f"{self.player.name} ({self.player.level})"

        with open(self.saveListFilePath, 'w') as outfile:  
            json.dump(saveList, outfile)


class ItemList:
  items = [
    {
      "id": 1,
      "kind": "weapon",
      "name": "Club",
      "level": 1,
      "ac": "",
      "atk": 1,
      "type": "blunt",
      "ability": "",
      "effect": ""
    },
    {
      "id": 2,
      "kind": "weapon",
      "name": "Hatchet",
      "level": 1,
      "ac": "",
      "atk": 1,
      "type": "slash",
      "ability": "",
      "effect": ""
    },
    {
      "id": 3,
      "kind": "weapon",
      "name": "Knife",
      "level": 1,
      "ac": "",
      "atk": 1,
      "type": "pierce",
      "ability": "",
      "effect": ""
    },
    {
      "id": 4,
      "kind": "weapon",
      "name": "Axe",
      "level": 2,
      "ac": "",
      "atk": 2,
      "type": "slash",
      "ability": "",
      "effect": ""
    },
    {
      "id": 5,
      "kind": "weapon",
      "name": "Staff",
      "level": 2,
      "ac": "",
      "atk": 2,
      "type": "blunt",
      "ability": "",
      "effect": ""
    },
    {
      "id": 6,
      "kind": "weapon",
      "name": "Dagger",
      "level": 2,
      "ac": "",
      "atk": 2,
      "type": "pierce",
      "ability": "",
      "effect": ""
    },
    {
      "id": 7,
      "kind": "weapon",
      "name": "Sword",
      "level": 3,
      "ac": "",
      "atk": 3,
      "type": "slash",
      "ability": "",
      "effect": ""
    },
    {
      "id": 8,
      "kind": "weapon",
      "name": "Mace",
      "level": 3,
      "ac": "",
      "atk": 3,
      "type": "blunt",
      "ability": "",
      "effect": ""
    },
    {
      "id": 9,
      "kind": "weapon",
      "name": "Spear",
      "level": 3,
      "ac": "",
      "atk": 3,
      "type": "pierce",
      "ability": "",
      "effect": ""
    },
    {
      "id": 10,
      "kind": "weapon",
      "name": "Claymore",
      "level": 4,
      "ac": "",
      "atk": 4,
      "type": "slash",
      "ability": "",
      "effect": ""
    },
    {
      "id": 11,
      "kind": "weapon",
      "name": "Warhammer",
      "level": 4,
      "ac": "",
      "atk": 4,
      "type": "blunt",
      "ability": "",
      "effect": ""
    },
    {
      "id": 12,
      "kind": "weapon",
      "name": "Rapier",
      "level": 4,
      "ac": "",
      "atk": 4,
      "type": "pierce",
      "ability": "",
      "effect": ""
    },
    {
      "id": 13,
      "kind": "weapon",
      "name": "Flamberge",
      "level": 5,
      "ac": "",
      "atk": 5,
      "type": "slash",
      "ability": "",
      "effect": ""
    },
    {
      "id": 14,
      "kind": "weapon",
      "name": "Morningstar",
      "level": 5,
      "ac": "",
      "atk": 5,
      "type": "blunt",
      "ability": "",
      "effect": ""
    },
    {
      "id": 15,
      "kind": "weapon",
      "name": "Pick",
      "level": 5,
      "ac": "",
      "atk": 5,
      "type": "pierce",
      "ability": "",
      "effect": ""
    },
    {
      "id": 16,
      "kind": "weapon",
      "name": "Glaive",
      "level": 6,
      "ac": "",
      "atk": 6,
      "type": "slash",
      "ability": "",
      "effect": ""
    },
    {
      "id": 17,
      "kind": "weapon",
      "name": "Flail",
      "level": 6,
      "ac": "",
      "atk": 6,
      "type": "blunt",
      "ability": "",
      "effect": ""
    },
    {
      "id": 18,
      "kind": "weapon",
      "name": "Pike",
      "level": 6,
      "ac": "",
      "atk": 6,
      "type": "pierce",
      "ability": "",
      "effect": ""
    },
    {
      "id": 19,
      "kind": "weapon",
      "name": "Halberd",
      "level": 7,
      "ac": "",
      "atk": 7,
      "type": "slash",
      "ability": "",
      "effect": ""
    },
    {
      "id": 20,
      "kind": "weapon",
      "name": "Maul",
      "level": 7,
      "ac": "",
      "atk": 7,
      "type": "blunt",
      "ability": "",
      "effect": ""
    },
    {
      "id": 21,
      "kind": "weapon",
      "name": "Trident",
      "level": 7,
      "ac": "",
      "atk": 7,
      "type": "pierce",
      "ability": "",
      "effect": ""
    },
    {
      "id": 22,
      "kind": "armor",
      "name": "Robe",
      "level": 1,
      "ac": 1,
      "atk": "",
      "type": "cloth",
      "ability": "",
      "effect": ""
    },
    {
      "id": 23,
      "kind": "armor",
      "name": "Cuirboulli",
      "level": 2,
      "ac": 2,
      "atk": "",
      "type": "leather",
      "ability": "",
      "effect": ""
    },
    {
      "id": 24,
      "kind": "armor",
      "name": "Hauberk",
      "level": 3,
      "ac": 3,
      "atk": "",
      "type": "metal",
      "ability": "",
      "effect": ""
    },
    {
      "id": 25,
      "kind": "armor",
      "name": "Jack",
      "level": 4,
      "ac": 4,
      "atk": "",
      "type": "leather",
      "ability": "",
      "effect": ""
    },
    {
      "id": 26,
      "kind": "armor",
      "name": "Brigandine",
      "level": 5,
      "ac": 5,
      "atk": "",
      "type": "metal",
      "ability": "",
      "effect": ""
    },
    {
      "id": 27,
      "kind": "armor",
      "name": "Breastplate",
      "level": 6,
      "ac": 6,
      "atk": "",
      "type": "metal",
      "ability": "",
      "effect": ""
    },
    {
      "id": 28,
      "kind": "armor",
      "name": "Cuirass",
      "level": 7,
      "ac": 7,
      "atk": "",
      "type": "metal",
      "ability": "",
      "effect": ""
    },
    {
      "id": 29,
      "kind": "ring",
      "name": "Ring of Shielding",
      "level": 3,
      "ac": 1,
      "atk": "",
      "type": "",
      "ability": "",
      "effect": ""
    },
    {
      "id": 30,
      "kind": "ring",
      "name": "Ring",
      "level": 2,
      "ac": "",
      "atk": "",
      "type": "",
      "ability": "traveling",
      "effect": ""
    },
    {
      "id": 31,
      "kind": "ring",
      "name": "Ring",
      "level": 3,
      "ac": "",
      "atk": "",
      "type": "",
      "ability": "running",
      "effect": ""
    },
    {
      "id": 32,
      "kind": "ring",
      "name": "Ring",
      "level": 4,
      "ac": "",
      "atk": "",
      "type": "",
      "ability": "regeneration",
      "effect": ""
    },
    {
      "id": 33,
      "kind": "usable",
      "name": "Potion",
      "level": 1,
      "ac": "",
      "atk": "",
      "type": "anytime",
      "ability": "",
      "effect": "healing"
    },
    {
      "id": 34,
      "kind": "usable",
      "name": "Scroll",
      "level": 1,
      "ac": "",
      "atk": "",
      "type": "peace",
      "ability": "",
      "effect": "digging"
    },
    {
      "id": 35,
      "kind": "usable",
      "name": "Scroll",
      "level": 1,
      "ac": "",
      "atk": "",
      "type": "anytime",
      "ability": "",
      "effect": "blinking"
    },
    {
      "id": 36,
      "kind": "usable",
      "name": "Scroll",
      "level": 2,
      "ac": "",
      "atk": "",
      "type": "peace",
      "ability": "",
      "effect": "teleport"
    },
    {
      "id": 37,
      "kind": "usable",
      "name": "Potion",
      "level": 2,
      "ac": "",
      "atk": "",
      "type": "combat",
      "ability": "",
      "effect": "invisibility"
    },
    {
      "id": 38,
      "kind": "usable",
      "name": "Scroll",
      "level": 2,
      "ac": "",
      "atk": "",
      "type": "peace",
      "ability": "",
      "effect": "shopping"
    },
    {
      "id": 39,
      "kind": "usable",
      "name": "Scroll",
      "level": 4,
      "ac": "",
      "atk": "",
      "type": "peace",
      "ability": "",
      "effect": "time warp"
    },
    {
      "id": 40,
      "kind": "usable",
      "name": "Scroll",
      "level": 3,
      "ac": "",
      "atk": "",
      "type": "combat",
      "ability": "",
      "effect": "fireball"
    },
    {
      "id": 41,
      "kind": "usable",
      "name": "Potion",
      "level": 4,
      "ac": "",
      "atk": "",
      "type": "peace",
      "ability": "",
      "effect": "knowledge"
    }
  ]

  descriptors = {
      "weapon": {
          "pierce": [
              "Light",
              "Balanced",
              "Tapered",
              "Pointy",
              "Sharp",
              "Keen",
              "Piercing",
              "Barbed",
              "Impaling",
              "Stabbing",
              "Needle",
              "Diamond"
          ],
          "slash": [
              "Fine",
              "Long",
              "Honed",
              "Jagged",
              "Razor",
              "Gleaming",
              "Serrated",
              "Tempered",
              "Curved",
              "Keen",
              "Double Edged",
              "Vorpal"
          ],
          "blunt": [
              "Polished",
              "Hardened",
              "Dense",
              "Faceted",
              "Iron",
              "Heavy",
              "Leaden",
              "Massive",
              "Dragonbone",
              "Crushing",
              "Giant",
              "Unstoppable"
          ]
      },
      "armor": {
          "cloth": [
              "Linen",
              "Thick",
              "Wool",
              "Striped",
              "Felted",
              "Patterned",
              "Silk",
              "Embroidered",
              "Brocade",
              "Velvet",
              "Goldworked",
              "Mageweave"
          ],
          "leather": [
              "Soft",
              "Deerskin",
              "Waxed",
              "Hardened",
              "Dyed",
              "Embossed",
              "Tooled",
              "Lacquered",
              "Rhino Hide",
              "Gilded",
              "Trollskin",
              "Dragonscale"
          ],
          "metal": [
              "Copper",
              "Bronze",
              "Shiny",
              "Iron",
              "Polished",
              "Steel",
              "Burnished",
              "Exotic",
              "Engraved",
              "Chased",
              "Mithril",
              "Adamantine"
          ]
      },
      "ring": [
          "Shiny",
          "Golden",
          "Amethyst",
          "Agate",
          "Lustrous",
          "Garnet",
          "Malachite",
          "Hematite",
          "Emerald",
          "Elvish",
          "Ruby",
          "Dwarvish",
          "Sapphire",
          "Starmetal",
          "Glowing",
          "Quicksilver",
          "Crystal",
          "Dragonbone",
        
      ]
  }

  atkVerb = {
      "slash": [
          "slash",
          "cut",
          "hack"
      ],
      "pierce": [
          "stab",
          "poke",
          "impale"
      ],
      "blunt": [
          "bash",
          "smash",
          "crush"
      ],
      "acid": [
          "dissolve",
          "burn",
          "melt"
      ],
      "fire": [
          "scorch",
          "burn",
          "char"
      ],
      "electric": [
          "shock",
          "zap",
          "jolt"
      ],
      "cold": [
          "freeze",
          "chill",
          "frost"
      ]
  }

  weaponEgo = [
    {
      "name": "Flame",
      "bonus": 1,
      "attributes": {
        "type": "fire"
      }
    },
    {
      "name": "Frost",
      "bonus": 1,
      "attributes": {
        "type": "cold"
      }
    },
    {
      "name": "Acid",
      "bonus": 1,
      "attributes": {
        "type": "acid"
      }
    },
    {
      "name": "Shock",
      "bonus": 1,
      "attributes": {
        "type": "electric"
      }
    },
    {
      "name": "Power",
      "bonus": 2,
      "attributes": {}
    }
  ]

  armorEgo = [
    {
      "name": "Cooling",
      "bonus": 1,
      "attributes": {
        "ability": "resist_fire"
      }
    },
    {
      "name": "Warmth",
      "bonus": 1,
      "attributes": {
        "ability": "resist_cold"
      }
    },
    {
      "name": "Alkali",
      "bonus": 1,
      "attributes": {
        "ability": "resist_acid"
      }
    },
    {
      "name": "Insulation",
      "bonus": 1,
      "attributes": {
        "ability": "resist_electric"
      }
    },
    {
      "name": "Health",
      "bonus": 0,
      "attributes": {
        "ability": "regeneration" 
      }
    },
    {
      "name": "Speed",
      "bonus": 0,
      "attributes": {
        "ability": "running" 
      }
    },
    {
      "name": "Defense",
      "bonus": 2,
      "attributes": {}
    }
  ]


import random

from .item import Item
from .utils import Utils

class Store:
    def __init__(self, level):
        self.level = level
        self.items = []
        self.maxItems = 30
        self.generateItems()

    def generateItems(self):
        while len(self.items) < self.maxItems:
            newItem = Item(random.randint(max(self.level - 2, 1), self.level))
            if newItem.kind:
                self.items += [newItem]

    def addItem(self, item):
        if kind == "usable":
            found = False
            for invItem in self.items:
                if invItem.id == item.id:
                    found = True
                    invItem.stack += 1
            if not found:
                self.items.append(item)
        else:
            self.items.append(item)

    def removeItem(self, item):
        if item.kind == "usable" and item.stack > 1:
            item.stack -= 1
        else:
            self.items.remove(item)


import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="thousandrooms",
    version="0.1.3",
    author="Ben Pung",
    author_email="ben@houseofpung.net",
    description="A simple roguelike RPG",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/bypung/thousandrooms",
    packages=setuptools.find_packages(),
    install_requires=[
        'colored',
        'colorama'
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    entry_points={
        'console_scripts': [
            'thousandrooms = thousandrooms.__main__:main'
        ]
    },
    include_package_data=True,
)